# coding: utf-8
# 2019/12/20 @ tongshiwei

"""
Liu Ding Shen Huo
"""
from ..analyser.select import get_max
from .hyper_search import extract_params_combinations


def get_validation_file(model):
    return model.mod.cfg.validation_result_file


class LDSH(object):
    def __init__(self, model, metric_key, *positional_params, fix_params=None):
        attributes = ["train"]
        for attr in attributes:
            assert hasattr(model, attr)

        self.elixir = model
        self.fix_params = fix_params if fix_params else {}
        self.positional_params = positional_params
        self.metric_key = metric_key

    def search(self, candidates):
        for params in extract_params_combinations(candidates, self.fix_params):
            _params = {}
            _params.update(params)
            _params.update(self.fix_params)
            module = self.elixir.train(
                *self.positional_params,
                **_params
            )
            self.get_metric()

    def get_metric(self):
        result, _ = get_max(get_validation_file(self.elixir), self.metric_key, )
        return result[self.metric_key]



# coding: utf-8
# 2019/12/20 @

import itertools
from longling import as_list
from longling.lib.candylib import dict2pv, list2dict, get_dict_by_path


def extract_params_combinations(candidates: dict, external=None):
    """
    >>> candidates = {'b': [1, 2], 'c': [0, 3], 'd': '$b'}
    >>> list(extract_params_combinations(candidates))
    [{'b': 1, 'c': 0, 'd': 1}, {'b': 1, 'c': 3, 'd': 1}, {'b': 2, 'c': 0, 'd': 2}, {'b': 2, 'c': 3, 'd': 2}]
    >>> candidates = {'a': [1, 2], 'b': '$c'}
    >>> external = {'c': 3}
    >>> list(extract_params_combinations(candidates, external))
    [{'a': 1, 'b': 3}, {'a': 2, 'b': 3}]
    """
    external = {} if external is None else external
    params_paths, params_values = dict2pv(candidates)
    params_values = [as_list(value) for value in params_values]

    for params in itertools.product(*params_values):
        _params = {}

        for p, v in zip(params_paths, params):
            list2dict(p, v, _params)

        for p, v in zip(params_paths, params):
            if isinstance(v, str) and v[0] == '$':
                map_key_path = v.lstrip('$').split(":")
                _dict_obj = get_dict_by_path(_params, p[:-1])
                for map_dict in [_params, external]:
                    try:
                        _map_dict_obj = get_dict_by_path(map_dict, map_key_path)
                        _dict_obj[p[-1]] = _map_dict_obj
                        break
                    except KeyError:
                        try:
                            _map_dict_obj = get_dict_by_path(map_dict, p[:-1] + map_key_path)
                            _dict_obj[p[-1]] = _map_dict_obj
                            break
                        except KeyError:
                            pass

                else:
                    raise KeyError(
                        "The mapped key should be in either candidates or external, but cannot find %s" % v
                    )

        yield _params



__all__ = ['as_list', 'dict2pv', 'list2dict', 'get_dict_by_path']


def as_list(obj):
    r"""A utility function that converts the argument to a list
    if it is not already.

    Parameters
    ----------
    obj : object
        argument to be converted to a list

    Returns
    -------
    list_obj: list
        If `obj` is a list or tuple, return it. Otherwise,
        return `[obj]` as a single-element list.

    Examples
    --------
    >>> as_list(1)
    [1]
    >>> as_list([1])
    [1]
    """
    if isinstance(obj, (list, tuple)):
        return obj
    else:
        return [obj]


def list2dict(list_obj, value=None, dict_obj=None):
    """
    >>> list_obj = ["a", 2, "c"]
    >>> list2dict(list_obj, 10)
    {'a': {2: {'c': 10}}}
    """
    dict_obj = {} if dict_obj is None else dict_obj
    _dict_obj = dict_obj
    for e in list_obj[:-1]:
        if e not in _dict_obj:
            _dict_obj[e] = {}
        _dict_obj = _dict_obj[e]
    _dict_obj[list_obj[-1]] = value

    return dict_obj


def get_dict_by_path(dict_obj, path_to_node):
    """
    >>> dict_obj = {"a": {"b": {"c": 1}}}
    >>> get_dict_by_path(dict_obj, ["a", "b", "c"])
    1
    """
    _dict_obj = dict_obj
    for p in path_to_node:
        _dict_obj = _dict_obj[p]
    return _dict_obj


def dict2pv(dict_obj: dict, path_to_node: list = None):
    """
    >>> dict_obj = {"a": {"b": [1, 2], "c": "d"}, "e": 1}
    >>> a, b = dict2pv(dict_obj)
    >>> a
    [['a', 'b'], ['a', 'c'], ['e']]
    >>> b
    [[1, 2], 'd', 1]
    """
    paths = []
    values = []
    path_to_node = [] if path_to_node is None else path_to_node
    for key, value in dict_obj.items():
        if isinstance(value, dict):
            _paths, _values = dict2pv(dict_obj[key], path_to_node + [key])
            paths.extend(_paths)
            values.extend(_values)
        else:
            paths.append(path_to_node + [key])
            values.append(value)
    return paths, values


def get_all_subclass(cls):  # pragma: no cover
    subclass = set()

    def _get_all_subclass(cls, res_set):
        res_set.add(cls)
        for sub_cls in cls.__subclasses__():
            _get_all_subclass(sub_cls, res_set)

    _get_all_subclass(cls, res_set=subclass)
    return subclass


class RegisterNameExistedError(Exception):  # pragma: no cover
    pass


class Register(object):  # pragma: no cover
    def __init__(self):
        self.reg = set()

    def register(self, name):
        if name not in self.reg:
            self.reg.add(name)
            return True
        else:
            raise RegisterNameExistedError("name %s existed" % name)


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# coding: utf-8
# 2019/12/24 @ tongshiwei


class IncrementEstimator(object):
    def __init__(self, step, init_value=None):
        self._value = init_value
        self._step = step

    def __call__(self, value):
        self._value = self._value + self._step * (value - self._value) if self._value else value
        return self.value

    @property
    def value(self):
        return self._value
